import pathlib
import unittest

import numpy as np

from model_server.conf.testing import czifile, ilastik_classifiers, output_path, roiset_test_data
from model_server.base.accessors import CziImageFileAccessor, generate_file_accessor, InMemoryDataAccessor, PatchStack, write_accessor_data_to_file
from model_server.extensions.ilastik import models as ilm
from model_server.extensions.ilastik.workflows import infer_px_then_ob_model
from model_server.base.models import InvalidObjectLabelsError
from model_server.base.roiset import _get_label_ids, RoiSet, RoiSetMetaParams
from model_server.base.workflows import classify_pixels
from tests.test_api import TestServerBaseClass

def _random_int(*args):
    return np.random.randint(0, 2 ** 8, size=args, dtype='uint8')

class TestIlastikPixelClassification(unittest.TestCase):
    def setUp(self) -> None:
        self.cf = CziImageFileAccessor(czifile['path'])
        self.channel = 0
        self.model = ilm.IlastikPixelClassifierModel(
            params=ilm.IlastikPixelClassifierParams(project_file=ilastik_classifiers['px'].__str__())
        )
        self.mono_image = self.cf.get_mono(self.channel)


    def test_faulthandler(self): # recreate error that is messing up ilastik
        import io
        import sys
        import faulthandler

        with self.assertRaises(io.UnsupportedOperation):
            faulthandler.enable(file=sys.stdout)


    def test_raise_error_if_autoload_disabled(self):
        model = ilm.IlastikPixelClassifierModel(
            params=ilm.IlastikPixelClassifierParams(project_file=ilastik_classifiers['px'].__str__()),
            autoload=False
        )
        w = 512
        h = 256

        input_img = InMemoryDataAccessor(data=np.random.rand(w, h, 1, 1))

        with self.assertRaises(AttributeError):
            mask = model.label_pixel_class(input_img)


    def test_run_pixel_classifier_on_random_data(self):
        w = 512
        h = 256

        input_img = InMemoryDataAccessor(data=np.random.rand(h, w, 1, 1))

        mask = self.model.label_pixel_class(input_img)
        self.assertEqual(mask.shape, (h, w, 1, 1))


    def test_run_pixel_classifier(self):
        self.assertEqual(self.mono_image.shape_dict['X'], czifile['w'])
        self.assertEqual(self.mono_image.shape_dict['Y'], czifile['h'])
        self.assertEqual(self.mono_image.shape_dict['C'], 1)
        self.assertEqual(self.mono_image.shape_dict['Z'], 1)

        mask = self.model.label_pixel_class(self.mono_image)

        self.assertTrue(mask.is_mask())
        self.assertEqual(mask.shape[0:2], self.cf.shape[0:2])
        self.assertEqual(mask.shape_dict['C'], 1)
        self.assertEqual(mask.shape_dict['Z'], 1)
        self.assertTrue(
            write_accessor_data_to_file(
                output_path / 'seg' / f'seg_{self.cf.fpath.stem}_ch{self.channel}.tif',
                mask
            )
        )

    def test_label_pixels_with_params(self):
        def _run_seg(tr, sig):
            mod = ilm.IlastikPixelClassifierModel(
                params=ilm.IlastikPixelClassifierParams(
                    project_file=ilastik_classifiers['px'].__str__(),
                    px_prob_threshold=tr,
                    px_smoothing=sig,
                ),
            )
            mask = mod.label_pixel_class(self.mono_image)
            write_accessor_data_to_file(
                output_path / 'seg' / f'seg_tr{int(10*tr)}_sig{int(10*sig)}.tif',
                mask
            )
            return mask
        mask1 = _run_seg(0.5, 0.0)
        mask2 = _run_seg(0.5, 0.2)
        self.assertEqual(mask1.shape, mask2.shape)


    def test_pixel_classifier_enforces_input_shape(self):
        self.assertEqual(self.model.model_chroma, 1)
        self.assertEqual(self.model.model_3d, False)

        # correct data
        self.assertIsInstance(
            self.model.label_pixel_class(
                InMemoryDataAccessor(
                    _random_int(512, 256, 1, 1)
                )
            ),
            InMemoryDataAccessor
        )

        # raise except with input of multiple channels
        with self.assertRaises(ilm.IlastikInputShapeError):
            mask = self.model.label_pixel_class(
                InMemoryDataAccessor(
                    _random_int(512, 256, 3, 1)
                )
            )

        # raise except with input of multiple channels
        with self.assertRaises(ilm.IlastikInputShapeError):
            mask = self.model.label_pixel_class(
                InMemoryDataAccessor(
                    _random_int(512, 256, 1, 15)
                )
            )

    def test_ilastik_infer_pxmap_from_patchstack(self):

        def _r(h):
            return np.random.randint(0, 2 ** 8, size=(h, 512, 1, 1), dtype='uint8')

        acc = PatchStack([_r(256), _r(512), _r(256)])
        self.assertEqual(acc.hw, (512, 512))
        self.assertEqual(acc.iat(0, crop=True).hw, (256, 512))

        mask = self.model.label_patch_stack(acc)
        self.assertEqual(mask.dtype, bool)
        self.assertEqual(mask.chroma, 1)
        self.assertEqual(mask.hw, acc.hw)
        self.assertEqual(mask.nz, acc.nz)
        self.assertEqual(mask.count, acc.count)

        pxmap, _ = self.model.infer_patch_stack(acc)
        self.assertEqual(pxmap.dtype, float)
        self.assertEqual(pxmap.chroma, len(self.model.labels))
        self.assertEqual(pxmap.hw, acc.hw)
        self.assertEqual(pxmap.nz, acc.nz)
        self.assertEqual(pxmap.count, acc.count)

    def test_run_object_classifier_from_pixel_predictions(self):
        self.test_run_pixel_classifier()
        fp = czifile['path']
        model = ilm.IlastikObjectClassifierFromPixelPredictionsModel(
            params=ilm.IlastikParams(project_file=ilastik_classifiers['pxmap_to_obj'].__str__())
        )
        mask = self.model.label_pixel_class(self.mono_image)
        objmap, _ = model.infer(self.mono_image, mask)

        self.assertTrue(
            write_accessor_data_to_file(
                output_path / f'obmap_{fp.stem}.tif',
                objmap,
            )
        )
        self.assertEqual(objmap.data.max(), 2)


    def test_run_object_classifier_from_segmentation(self):
        self.test_run_pixel_classifier()
        fp = czifile['path']
        model = ilm.IlastikObjectClassifierFromSegmentationModel(
            params=ilm.IlastikParams(project_file=ilastik_classifiers['seg_to_obj'].__str__())
        )
        mask = self.model.label_pixel_class(self.mono_image)
        objmap = model.label_instance_class(self.mono_image, mask)

        self.assertTrue(
            write_accessor_data_to_file(
                output_path / f'obmap_from_seg_{fp.stem}.tif',
                objmap,
            )
        )
        self.assertEqual(objmap.data.max(), 2)

    def test_ilastik_pixel_classification_as_workflow(self):
        result = classify_pixels(
            czifile['path'],
            ilm.IlastikPixelClassifierModel(
                params=ilm.IlastikPixelClassifierParams(project_file=ilastik_classifiers['px'].__str__()),
            ),
            output_path,
            channel=0,
        )
        self.assertTrue(result.success)
        self.assertGreater(result.timer_results['inference'], 0.1)

class TestIlastikOverApi(TestServerBaseClass):

    def test_httpexception_if_incorrect_project_file_loaded(self):
        resp_load = self._put(
            'ilastik/seg/load/',
            body={'project_file': 'improper.ilp'},
        )
        self.assertEqual(resp_load.status_code, 500)


    def test_load_ilastik_pixel_model(self):
        resp_load = self._put(
            'ilastik/seg/load/',
            body={'project_file': str(ilastik_classifiers['px'])},
        )
        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        model_id = resp_load.json()['model_id']
        resp_list = self._get('models')
        self.assertEqual(resp_list.status_code, 200)
        rj = resp_list.json()
        self.assertEqual(rj[model_id]['class'], 'IlastikPixelClassifierModel')
        return model_id

    def test_load_another_ilastik_pixel_model(self):
        model_id = self.test_load_ilastik_pixel_model()
        resp_list_1st = self._get('models').json()
        self.assertEqual(len(resp_list_1st), 1, resp_list_1st)
        resp_load_2nd = self._put(
            'ilastik/seg/load/',
            body={'project_file': str(ilastik_classifiers['px']), 'duplicate': True},
        )
        resp_list_2nd = self._get('models').json()
        self.assertEqual(len(resp_list_2nd), 2, resp_list_2nd)
        resp_load_3rd = self._put(
            'ilastik/seg/load/',
            body={'project_file': str(ilastik_classifiers['px']), 'duplicate': False},
        )
        resp_list_3rd = self._get('models').json()
        self.assertEqual(len(resp_list_3rd), 2, resp_list_3rd)

    def test_load_ilastik_pixel_model_with_params(self):
        params = {
            'project_file': str(ilastik_classifiers['px']),
            'px_class': 0,
            'px_prob_threshold': 0.5
        }
        resp_load = self._put(
            'ilastik/seg/load/',
            body=params,
        )
        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        model_id = resp_load.json()['model_id']
        mods = self._get('models').json()
        self.assertEqual(len(mods), 1)
        self.assertEqual(mods[model_id]['params']['px_prob_threshold'], 0.5)


    def test_load_ilastik_pxmap_to_obj_model(self):
        resp_load = self._put(
            'ilastik/pxmap_to_obj/load/',
            body={'project_file': str(ilastik_classifiers['pxmap_to_obj'])},
        )
        model_id = resp_load.json()['model_id']

        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        resp_list = self._get('models')
        self.assertEqual(resp_list.status_code, 200)
        rj = resp_list.json()
        self.assertEqual(rj[model_id]['class'], 'IlastikObjectClassifierFromPixelPredictionsModel')
        return model_id

    def test_load_ilastik_seg_to_obj_model(self):
        resp_load = self._put(
            'ilastik/seg_to_obj/load/',
            body={'project_file': str(ilastik_classifiers['seg_to_obj'])},
        )
        model_id = resp_load.json()['model_id']

        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        resp_list = self._get('models')
        self.assertEqual(resp_list.status_code, 200)
        rj = resp_list.json()
        self.assertEqual(rj[model_id]['class'], 'IlastikObjectClassifierFromSegmentationModel')
        return model_id

    def test_ilastik_infer_pixel_probability(self):
        self.copy_input_file_to_server()
        model_id = self.test_load_ilastik_pixel_model()

        resp_infer = self._put(
            f'workflows/segment',
            query={'model_id': model_id, 'input_filename': czifile['filename'], 'channel': 0},
        )
        self.assertEqual(resp_infer.status_code, 200, resp_infer.content.decode())


    def test_ilastik_infer_px_then_ob(self):
        self.copy_input_file_to_server()
        px_model_id = self.test_load_ilastik_pixel_model()
        ob_model_id = self.test_load_ilastik_pxmap_to_obj_model()

        resp_infer = self._put(
            'ilastik/pixel_then_object_classification/infer/',
            query={
                'px_model_id': px_model_id,
                'ob_model_id': ob_model_id,
                'input_filename': czifile['filename'],
                'channel': 0,
            }
        )
        self.assertEqual(resp_infer.status_code, 200, resp_infer.content.decode())


class TestIlastikOnMultichannelInputs(TestServerBaseClass):
    def setUp(self) -> None:
        super(TestIlastikOnMultichannelInputs, self).setUp()
        self.pa_px_classifier = ilastik_classifiers['px_color_zstack']
        self.pa_ob_pxmap_classifier = ilastik_classifiers['ob_pxmap_color_zstack']
        self.pa_ob_seg_classifier = ilastik_classifiers['ob_seg_color_zstack']
        self.pa_input_image = roiset_test_data['multichannel_zstack']['path']
        self.pa_mask = roiset_test_data['multichannel_zstack']['mask_path_3d']

    def _copy_input_file_to_server(self):
        from shutil import copyfile

        pa_data = roiset_test_data['multichannel_zstack']['path']
        resp = self._get('paths')
        pa = resp.json()['inbound_images']
        outpath = pathlib.Path(pa) / pa_data.name
        copyfile(
            czifile['path'],
            outpath
        )

    def test_classify_pixels(self):
        img = generate_file_accessor(self.pa_input_image)
        self.assertGreater(img.chroma, 1)
        mod = ilm.IlastikPixelClassifierModel(ilm.IlastikPixelClassifierParams(project_file=self.pa_px_classifier.__str__()))
        pxmap = mod.infer(img)[0]
        self.assertEqual(pxmap.hw, img.hw)
        self.assertEqual(pxmap.nz, img.nz)
        return pxmap

    def test_classify_objects(self):
        pxmap = self.test_classify_pixels()
        img = generate_file_accessor(self.pa_input_image)
        mod = ilm.IlastikObjectClassifierFromPixelPredictionsModel(
            ilm.IlastikParams(project_file=self.pa_ob_pxmap_classifier.__str__())
        )
        obmap = mod.infer(img, pxmap)[0]
        self.assertEqual(obmap.hw, img.hw)
        self.assertEqual(obmap.nz, img.nz)

    def _call_workflow(self, channel):
        return infer_px_then_ob_model(
            self.pa_input_image,
            ilm.IlastikPixelClassifierModel(
                ilm.IlastikParams(project_file=self.pa_px_classifier.__str__()),
            ),
            ilm.IlastikObjectClassifierFromPixelPredictionsModel(
                ilm.IlastikParams(project_file=self.pa_ob_pxmap_classifier.__str__()),
            ),
            output_path,
            channel=channel,
        )

    def test_workflow(self):
        with self.assertRaises(ilm.IlastikInputShapeError):
            self._call_workflow(channel=0)
        res = self._call_workflow(channel=None)
        acc_input = generate_file_accessor(self.pa_input_image)
        acc_obmap = generate_file_accessor(res.object_map_filepath)
        self.assertEqual(acc_obmap.hw, acc_input.hw)
        self.assertEqual(len(acc_obmap._unique()[1]), 3)


    def test_api(self):
        resp_load = self._put(
            'ilastik/seg/load/',
            body={'project_file': str(self.pa_px_classifier)},
        )
        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        px_model_id = resp_load.json()['model_id']

        resp_load = self._put(
            'ilastik/pxmap_to_obj/load/',
            body={'project_file': str(self.pa_ob_pxmap_classifier)},
        )
        self.assertEqual(resp_load.status_code, 200, resp_load.json())
        ob_model_id = resp_load.json()['model_id']

        resp_infer = self._put(
            'ilastik/pixel_then_object_classification/infer/',
            query={
                'px_model_id': px_model_id,
                'ob_model_id': ob_model_id,
                'input_filename': self.pa_input_image.__str__(),
            }
        )
        self.assertEqual(resp_infer.status_code, 200, resp_infer.content.decode())
        acc_input = generate_file_accessor(self.pa_input_image)
        acc_obmap = generate_file_accessor(resp_infer.json()['object_map_filepath'])
        self.assertEqual(acc_obmap.hw, acc_input.hw)


class TestIlastikObjectClassification(unittest.TestCase):
    def setUp(self):
        stack = generate_file_accessor(roiset_test_data['multichannel_zstack']['path'])
        stack_ch_pa = stack.get_mono(roiset_test_data['pipeline_params']['patches_channel'])
        seg_mask = generate_file_accessor(roiset_test_data['multichannel_zstack']['mask_path'])

        self.roiset = RoiSet(
            stack_ch_pa,
            _get_label_ids(seg_mask),
            params=RoiSetMetaParams(
                mask_type='boxes',
                filters={'area': {'min': 1e3, 'max': 1e4}},
                expand_box_by=(64, 2)
            )
        )

        self.classifier = ilm.IlastikObjectClassifierFromSegmentationModel(
            params=ilm.IlastikParams(project_file=ilastik_classifiers['seg_to_obj'].__str__()),
        )
        self.raw = self.roiset.get_patches_acc()
        self.masks = self.roiset.get_patch_masks_acc()


    def test_classify_patches(self):
        res = self.classifier.label_patch_stack(self.raw, self.masks)
        self.assertEqual(res.count, self.roiset.count)
        res.export_pyxcz(output_path / 'res_patches.tif')
        for pi in range(0, res.count):  # assert that there is only one nonzero label per patch
            la, ct = np.unique(res.iat(pi).data, return_counts=True)
            self.assertEqual(np.sum(ct > 1), 2)  # exclude single-pixel anomaly
            self.assertEqual(la[0], 0)